﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Windows.Forms;

namespace LargeAddressAware
{
    public delegate void ThreadAddFileCompleteHandler();

    public partial class Main : Form
    {
        public event ThreadAddFileCompleteHandler ThreadAddFileCompleted;

        private List<LaaFile> _Files = new List<LaaFile>();
        private Dictionary<string, bool> _Originals = new Dictionary<string, bool>();
        private ListViewColumnSorter _ColumnSorter = new ListViewColumnSorter();

        private const string EXT_PAT_LONG = "Executables (*.exe)|*.exe";
        private const string EXT_PAT_SHORT = "*.exe";
        private const string EXT_PAT = ".exe";
        private readonly string SETTINGS_PATH = Path.GetDirectoryName(Assembly.GetExecutingAssembly().GetName().CodeBase).Substring(6) + @"\settings.sav";

        private AppMode _Mode = AppMode.Basic;
        private Size _Size = new Size(600, 300);

        private List<Thread> _Threads = new List<Thread>();
        private void RemoveRetiredThreads()
        {
            for (int i = _Threads.Count - 1; i >= 0; i--)
            {
                if ((_Threads[i].ThreadState == ThreadState.Stopped) || (_Threads[i].ThreadState == ThreadState.Aborted))
                    _Threads.RemoveAt(i);
            }
        }
        private void AbortThreads()
        {
            for (int i = _Threads.Count - 1; i >= 0; i--)
                _Threads[i].Abort();
        }

        private string _DefaultDirectory = "";

        public Main(string[] args)
        {
            InitializeComponent();
            lvFiles.ListViewItemSorter = _ColumnSorter;

            if (File.Exists(SETTINGS_PATH))
            {
                try
                {
                    FileStream fs = new FileStream(SETTINGS_PATH, FileMode.Open, FileAccess.Read);
                    BinaryReader br = new BinaryReader(fs);
                    _Mode = (AppMode)br.ReadByte();
                    _Size.Width = br.ReadInt32();
                    _Size.Height = br.ReadInt32();
                    int count = br.ReadInt32();
                    string temp;

                    for (int i = 0; i < count; i++)
                    {
                        temp = br.ReadString();
                        _Originals.Add(temp, br.ReadBoolean());
                    }

                    if (fs.Position != fs.Length)
                    {
                        _DefaultDirectory = br.ReadString();

                        if (fs.Position != fs.Length)
                            mmOptionsLoadPreviousFiles.Checked = br.ReadBoolean();
                    }

                    br.Close();
                    fs.Close();
                }
                catch { File.Delete(SETTINGS_PATH); }
            }

            foreach (string file in args)
            {
                if (File.Exists(file))
                    AddFile(file);
            }

            if (mmOptionsLoadPreviousFiles.Checked)
                LoadPreviousFiles();

            // Switch to the appropriate mode.
            switch (_Mode)
            {
                case AppMode.Advanced:
                    SwitchMode(mmModeAdvanced, new EventArgs());
                    break;
                case AppMode.Intermediate:
                    SwitchMode(mmModeIntermediate, new EventArgs());
                    break;
                default:
                    if (_Files.Count > 1) // Basic can't handle more than one.
                        SwitchMode(mmModeIntermediate, new EventArgs());
                    else
                    {
                        if (mmOptionsLoadPreviousFiles.Checked)
                            SwitchMode(mmModeIntermediate, new EventArgs());
                        else
                            SwitchMode(mmModeBasic, new EventArgs());
                    }
                    break;
            }

            ThreadAddFileCompleted += new ThreadAddFileCompleteHandler(ThreadAddFileComplete);

            RefreshFiles();
        }
        private void Main_FormClosing(object sender, FormClosingEventArgs e)
        {
            AbortThreads();
            RemoveRetiredThreads();

            if (_Originals.Count > 0)
            {
                FileStream fs = new FileStream(SETTINGS_PATH, FileMode.Create, FileAccess.Write);
                BinaryWriter bw = new BinaryWriter(fs);
                bw.Write((byte)_Mode);
                bw.Write(_Size.Width);
                bw.Write(_Size.Height);
                bw.Write(_Originals.Count);
                foreach (string key in _Originals.Keys)
                {
                    bw.Write(key);
                    bw.Write(_Originals[key]);
                    bw.Flush();
                }
                bw.Write(_DefaultDirectory);
                bw.Write(mmOptionsLoadPreviousFiles.Checked);
                bw.Flush();

                bw.Close();
                fs.Close();
            }
        }

        private void SortColumn(object sender, ColumnClickEventArgs e)
        {
            // Determine if clicked column is already the column that is being sorted.
            if (e.Column == _ColumnSorter.SortColumn)
            {
                // Reverse the current sort direction for this column.
                if (_ColumnSorter.Order == SortOrder.Ascending)
                    _ColumnSorter.Order = SortOrder.Descending;
                else
                    _ColumnSorter.Order = SortOrder.Ascending;
            }
            else
            {
                // Set the column number that is to be sorted; default to ascending.
                _ColumnSorter.SortColumn = e.Column;
                _ColumnSorter.Order = SortOrder.Ascending;
            }

            // Perform the sort with these new sort options.
            ListView s = (ListView)sender;
            s.Sort();
        }
        private void SwitchMode(object sender, EventArgs e)
        {
            ToolStripDropDownItem cast = (ToolStripDropDownItem)sender;
            switch (cast.Text)
            {
                case "Advanced":
                    _Mode = AppMode.Advanced;

                    mmAdd.Visible = true;
                    mmSelect.Visible = true;
                    mmDo.Visible = true;
                    mmRemove.Visible = true;
                    mmIAddFiles.Visible = false;
                    mmISelectAll.Visible = false;
                    mmISwitchLaa.Visible = false;
                    lvFiles.Visible = true;
                    tlpMain.Visible = false;

                    this.MinimumSize = new Size(600, 300);
                    this.MaximumSize = new Size(0, 0);
                    this.Size = _Size;

                    RefreshFiles();
                    break;
                case "Intermediate":
                    _Mode = AppMode.Intermediate;

                    mmAdd.Visible = false;
                    mmSelect.Visible = false;
                    mmDo.Visible = false;
                    mmRemove.Visible = false;
                    mmIAddFiles.Visible = true;
                    mmISelectAll.Visible = true;
                    mmISwitchLaa.Visible = true;
                    lvFiles.Visible = true;
                    tlpMain.Visible = false;

                    this.MinimumSize = new Size(600, 300);
                    this.MaximumSize = new Size(0, 0);
                    this.Size = _Size;

                    RefreshFiles();
                    break;
                default:
                    _Mode = AppMode.Basic;

                    mmAdd.Visible = false;
                    mmSelect.Visible = false;
                    mmDo.Visible = false;
                    mmRemove.Visible = false;
                    mmIAddFiles.Visible = false;
                    mmISelectAll.Visible = false;
                    mmISwitchLaa.Visible = false;
                    lvFiles.Visible = false;
                    tlpMain.Visible = true;

                    this.MinimumSize = new Size(450, 250);
                    this.MaximumSize = new Size(450, 250);
                    this.Size = new Size(450, 250);

                    if (_Files.Count > 0)
                        txtPath.Text = _Files[0].Path;
                    break;
            }
        }

        private void RefreshFiles()
        {
            lvFiles.Items.Clear();
            for (int i = 0; i < _Files.Count; i++)
            {
                int index;
                try { index = lvFiles.Items.IndexOfKey(_Files[i].Path); }
                catch { index = -1; }

                bool original = _Files[i].LargeAddressAware;
                try { original = _Originals[_Files[i].Path]; }
                catch { _Originals.Add(_Files[i].Path, original); }

                if (index == -1)
                    lvFiles.Items.Add(new ListViewItem(new string[] { _Files[i].Path, _Files[i].LargeAddressAware.ToString(), original.ToString() }));
                else
                {
                    lvFiles.Items[index].SubItems[1].Text = _Files[i].LargeAddressAware.ToString();
                    lvFiles.Items[index].SubItems[2].Text = original.ToString();
                }
            }

            RemoveRetiredThreads();
        }
        private int IndexOfFile(string path)
        {
            int index = -1;
            for (int i = 0; i < _Files.Count; i++)
            {
                if (_Files[i].Path == path)
                {
                    index = i;
                    break;
                }
            }
            return index;
        }

        #region " Add "
        private void ThreadAddFileComplete()
        {
            if (this.InvokeRequired)
            {
                try { this.Invoke(new ThreadAddFileCompleteHandler(ThreadAddFileComplete)); }
                catch { }
            }
            else
                RefreshFiles();
        }
        private void ThreadAddFile(object path)
        {
            LaaFile file = new LaaFile((string)path);
            if (IndexOfFile(file.Path) == -1)
            {
                _Files.Add(file);
                ThreadAddFileCompleteHandler h = ThreadAddFileCompleted;
                if (h != null)
                    h();
            }
        }
        private void AddFile(string path)
        {
            if (IndexOfFile(path) == -1)
            {
                Thread t = new Thread(ThreadAddFile);
                t.Start(path);
                _Threads.Add(t);
            }
        }
        private void AddFiles(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.AddExtension = false;
            dlg.AutoUpgradeEnabled = true;
            dlg.CheckFileExists = true;
            dlg.CheckPathExists = true;
            dlg.DefaultExt = EXT_PAT;
            dlg.DereferenceLinks = true;
            dlg.Filter = EXT_PAT_LONG;
            dlg.Multiselect = true;
            dlg.ShowDialog();
            for (int i = 0; i < dlg.FileNames.Length; i++)
                AddFile(dlg.FileNames[i]);
        }
        private void AddFilesFolder(object sender, EventArgs e)
        {
            FolderBrowserDialog dlg = new FolderBrowserDialog();
            dlg.ShowNewFolderButton = false;
            dlg.SelectedPath = _DefaultDirectory;
            DialogResult result = dlg.ShowDialog();
            int count = _Files.Count;
            if ((dlg.SelectedPath.Length > 0) && (result == DialogResult.OK))
            {
                _DefaultDirectory = dlg.SelectedPath;
                foreach (string file in Directory.GetFiles(dlg.SelectedPath, EXT_PAT_SHORT))
                    AddFile(file);
            }
        }
        private void AddFilesFolderRecursive(object sender, EventArgs e)
        {
            FolderBrowserDialog dlg = new FolderBrowserDialog();
            dlg.ShowNewFolderButton = false;
            dlg.SelectedPath = _DefaultDirectory;
            DialogResult result = dlg.ShowDialog();
            int count = _Files.Count;
            if ((dlg.SelectedPath.Length > 0) && (result == DialogResult.OK))
            {
                _DefaultDirectory = dlg.SelectedPath;
                HarvestDirectory(dlg.SelectedPath);
            }
        }
        private void HarvestDirectory(string path)
        {
            foreach (string dir in Directory.GetDirectories(path))
                HarvestDirectory(dir);

            foreach (string file in Directory.GetFiles(path, EXT_PAT_SHORT))
                AddFile(file);
        }
        #endregion

        #region " Select "
        private void SelectAll(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
                lvFiles.Items[i].Checked = true;
        }
        private void SelectNone(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
                lvFiles.Items[i].Checked = false;
        }
        private void SelectInvert(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
                lvFiles.Items[i].Checked = !lvFiles.Items[i].Checked;
        }

        private void SelectLaa(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].SubItems[1].Text == "True")
                    lvFiles.Items[i].Checked = true;
                else
                    lvFiles.Items[i].Checked = false;
            }
        }
        private void SelectNonLaa(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].SubItems[1].Text == "False")
                    lvFiles.Items[i].Checked = true;
                else
                    lvFiles.Items[i].Checked = false;
            }
        }
        private void SelectOriginal(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].SubItems[1].Text == lvFiles.Items[i].SubItems[2].Text)
                    lvFiles.Items[i].Checked = true;
                else
                    lvFiles.Items[i].Checked = false;
            }
        }
        private void SelectNonOriginal(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].SubItems[1].Text != lvFiles.Items[i].SubItems[2].Text)
                    lvFiles.Items[i].Checked = true;
                else
                    lvFiles.Items[i].Checked = false;
            }
        }
        #endregion

        #region " With Selected "
        private void DoForceLaa(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].Checked)
                {
                    int index = IndexOfFile(lvFiles.Items[i].Text);
                    if (_Files[index].LargeAddressAware == false)
                        _Files[index].WriteCharacteristics(true);
                }
            }

            RefreshFiles();
        }
        private void DoRemoveLaa(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].Checked)
                {
                    int index = IndexOfFile(lvFiles.Items[i].Text);
                    if (_Files[index].LargeAddressAware == true)
                        _Files[index].WriteCharacteristics(false);
                }
            }

            RefreshFiles();
        }
        private void DoSwitchLaa(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].Checked)
                {
                    int index = IndexOfFile(lvFiles.Items[i].Text);
                    if (_Files[index].LargeAddressAware == true)
                        _Files[index].WriteCharacteristics(false);
                    else
                        _Files[index].WriteCharacteristics(true);
                }
            }

            RefreshFiles();
        }
        private void DoSetOriginal(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].Checked)
                {
                    if (lvFiles.Items[i].SubItems[1].Text != lvFiles.Items[i].SubItems[2].Text)
                    {
                        int index = IndexOfFile(lvFiles.Items[i].Text);

                        if (lvFiles.Items[i].SubItems[2].Text == "True")
                            _Files[index].WriteCharacteristics(true);
                        else
                            _Files[index].WriteCharacteristics(false);
                    }
                }
            }

            RefreshFiles();
        }
        private void DoSetOppositeOriginal(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].Checked)
                {
                    if (lvFiles.Items[i].SubItems[1].Text == lvFiles.Items[i].SubItems[2].Text)
                    {
                        int index = IndexOfFile(lvFiles.Items[i].Text);

                        if (lvFiles.Items[i].SubItems[2].Text == "True")
                            _Files[index].WriteCharacteristics(false);
                        else
                            _Files[index].WriteCharacteristics(true);
                    }
                }
            }

            RefreshFiles();
        }

        private void DoSetOriginalTrue(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].Checked)
                    _Originals[lvFiles.Items[i].Text] = true;
            }

            RefreshFiles();
        }
        private void DoSetOriginalFalse(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].Checked)
                    _Originals[lvFiles.Items[i].Text] = false;
            }

            RefreshFiles();
        }
        #endregion

        #region " Remove "
        private void RemoveAll(object sender, EventArgs e)
        {
            _Files.Clear();
            RefreshFiles();
        }
        private void RemoveSelected(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].Checked)
                {
                    int index = IndexOfFile(lvFiles.Items[i].Text);
                    _Files.RemoveAt(index);
                }
            }

            RefreshFiles();
        }

        private void mmRemoveCompletelySelected_Click(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                if (lvFiles.Items[i].Checked)
                {
                    int index = IndexOfFile(lvFiles.Items[i].Text);
                    _Originals.Remove(_Files[index].Path);
                    _Files.RemoveAt(index);
                }
            }

            RefreshFiles();

        }
        private void mmRemoveCompletelyAll_Click(object sender, EventArgs e)
        {
            for (int i = 0; i < lvFiles.Items.Count; i++)
            {
                int index = IndexOfFile(lvFiles.Items[i].Text);
                _Originals.Remove(_Files[index].Path);
                _Files.RemoveAt(index);
            }
            RefreshFiles();
        }
        #endregion

        #region " Basic "
        private void btnBrowse_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.AddExtension = false;
            dlg.AutoUpgradeEnabled = true;
            dlg.CheckFileExists = true;
            dlg.CheckPathExists = true;
            dlg.DefaultExt = EXT_PAT;
            dlg.DereferenceLinks = true;
            dlg.Filter = EXT_PAT_LONG;
            dlg.Multiselect = false;
            dlg.ShowDialog();
            if (dlg.FileName.Length > 0)
                txtPath.Text = dlg.FileName;
        }
        private void txtPath_TextChanged(object sender, EventArgs e)
        {
            if (File.Exists(txtPath.Text))
            {
                _Files.Clear();
                AddFile(txtPath.Text);

                while (_Files.Count == 0)
                    Thread.Sleep(100);

                chkLargeAddressAware.Checked = _Files[0].LargeAddressAware;
                chkLargeAddressAware.Enabled = true;
                btnSave.Enabled = true;

                lblStatus.Text = "The file has been loaded.  Check or uncheck the box then click on save.";
            }
            else
            {
                chkLargeAddressAware.Enabled = false;
                btnSave.Enabled = false;
            }
        }
        private void btnSave_Click(object sender, EventArgs e)
        {
            if (_Files.Count > 0)
            {
                _Files[0].WriteCharacteristics(chkLargeAddressAware.Checked);
                lblStatus.Text = "The update has been performed successfully.";
                txtPath.Text = "";
                _Files.Clear();
            }
        }
        #endregion

        #region " Drag and Drop "
        private new void DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop, false) == true)
                e.Effect = DragDropEffects.All;
        }
        private void lvFiles_DragDrop(object sender, DragEventArgs e)
        {
            string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);

            int count = _Files.Count;
            foreach (string file in files)
            {
                if (file.EndsWith(EXT_PAT))
                    AddFile(file);
                else
                    count--;
            }

            if (count != _Files.Count)
                RefreshFiles();
        }
        private void txtPath_DragDrop(object sender, DragEventArgs e)
        {
            string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);

            if (files.Length > 0)
                txtPath.Text = files[0];
        }
        #endregion

        #region " Options "
        private void mmOptionsLoadPreviousFiles_Click(object sender, EventArgs e)
        {
            if (mmOptionsLoadPreviousFiles.Checked)
            {
                mmModeBasic.Enabled = true;
                mmOptionsLoadPreviousFiles.Checked = false;

            }
            else
            {
                if (_Mode == AppMode.Basic)
                    SwitchMode(mmModeIntermediate, new EventArgs());

                mmModeBasic.Enabled = false;
                mmOptionsLoadPreviousFiles.Checked = true;

                LoadPreviousFiles();
            }
        }
        private void LoadPreviousFiles()
        {
            foreach (KeyValuePair<string, bool> pair in _Originals)
            {
                if (File.Exists(pair.Key))
                    AddFile(pair.Key);
            }
        }
        #endregion

    }
}
